import java.io.*;
import java.lang.Math;

/******************************************************************************
 * Manager krzywych w enginie. Singleton. Po utworzeniu laduje definicje
 * krzywych z pliku konfiguracyjnego. Mozna ich potem dowolnie uzywac.
 * Mozna tez dodawac nowe krzywe i zapisywac je.
 */
public class CurveManager
{
//--------------------------------------------------------------------------
// protected members
    
    protected static java.util.Vector s_curves = new java.util.Vector();
	protected static String s_curvefilename = "Data\\curve.dat";

	protected static CurveManager instance;

    protected CurveManager()
    {
        //read in curve file
		DataInputStream s = null;
		try
		{
			s = new DataInputStream(new FileInputStream(s_curvefilename));
		} catch (FileNotFoundException e)
		{
			GameObject.CrashLog("Curve definition file "+s_curvefilename+" not found\n");
			return;
		}
		try 
		{
            //GameObject.LogLine("loading");
			Curve c = null;
			
			while ( s.available() > 0 )
			{
				//read definition of curve
                //GameObject.LogLine("new");
                byte b;
				switch ( b=s.readByte() )
				{
				case Curve.CLASS_SPLINE : c = new CurveSpline(); break;
                //case Curve.CLASS_CAR_SETTING_SPLINE: c = new CurveCarSettingsSpline(); break;
                default: GameObject.LogLine("[CurveManager] WARNING: Unknown curve type "+b);
				}

                //GameObject.LogLine("curve "+c);

				c.name = s.readUTF();
				c.samples = s.readUnsignedShort();
				c.minx = s.readFloat();
				c.maxx = s.readFloat();
                c.miny = s.readFloat();
                c.maxy = s.readFloat();

				if ( (c.minx > c.maxx) || (c.miny > c.maxy) )
				{
					GameObject.CrashLog("Error reading curve data: minx >= maxx");
					return;
				}
				c.Read(s);

				s_curves.add(c);
				
				c.Build(c.samples);

                GameObject.LogLine("read curve definition: "+c.name+"["+c.toString()+"]");
			}

			s.close();

		}
		catch (IOException e)
		{
			GameObject.CrashLog("Error reading file: "+s_curvefilename+"\n");
			return;
		}
    }

//--------------------------------------------------------------------------
// public members

	//constructor of the singleton
	public static CurveManager GetInstance()
	{
		if (instance == null) instance = new CurveManager();
		return instance;
	}

	//writes all curves
	void Save()
	{
		DataOutputStream s = null;
		try
		{
			s = new DataOutputStream(new FileOutputStream(s_curvefilename));
		} catch (FileNotFoundException e)
		{
			GameObject.CrashLog("Curve definition file "+s_curvefilename+" not created\n");
			return;
		}
		try 
		{
			for (int i=0; i < s_curves.size(); i++)
			{
				Curve c = (Curve)s_curves.get(i);

                s.writeByte( c.GetType() );
				s.writeUTF( c.name );
				s.writeShort( c.samples );
				s.writeFloat( c.minx );
				s.writeFloat( c.maxx );
                s.writeFloat( c.miny );
                s.writeFloat( c.maxy );

				c.Write(s);
                
                GameObject.LogLine("saved curve definition: "+c.name+"["+c.toString()+"]");
			}

			s.close();

		}
		catch (IOException e)
		{
			GameObject.CrashLog("Error writing file: "+s_curvefilename+"\n");
			return;
		}
	}
    
    //gets curve index
    int GetCurveId(String name)
    {
        for (int i=0; i < s_curves.size(); i++)
        {
            if ( ((Curve)s_curves.get(i)).name.compareTo(name) == 0 )
                return i;
        }
        return -1;
    }

    //gets curve having its name
    Curve GetCurve(String name)
    {
        Curve found = GetCurve(GetCurveId(name));
        if( found == null )
            GameObject.CrashLog( "Can't find curve named <" + name + ">\n" );
        return found;
    }

    //gets curve having its index
    Curve GetCurve(int i)
    {
        if ( (i < 0) || (i > s_curves.size()-1)) 
            return null;

        return (Curve)s_curves.get(i);
    }
    
    int GetCurvesCount()
    {
        return s_curves.size();
    }
    
    void AddCurve(Curve curve)
    {
        //check for existng curves
        for (int i=0; i<s_curves.size(); i++)
        {
            if (curve == s_curves.elementAt(i))
                return;
        }
        
        //insert
        s_curves.add(curve);
        
    }
    
    void DeleteCurve(int idx)
    {
        if ( (idx < 0) || (idx > s_curves.size()) ) return;

        s_curves.removeElementAt(idx);
    }

    void Rebuild()
    {
        for (int i=0; i<s_curves.size(); i++)
        {
            Curve c = (Curve)s_curves.elementAt(i);
            c.Build(c.samples);
        }
    }

//--------------------------------------------------------------------------
// protected members
    
    protected java.util.Vector GetCurves()
    {
        return s_curves;
    }

}

/******************************************************************************
 * Bazowa klasa krzywych
 */
class Curve
{

	public final static byte CLASS_NONE = 0;
	public final static byte CLASS_SPLINE = 1;
    public final static byte CLASS_CAR_SETTING_SPLINE = 2;
    public final static byte CLASS_DEBUG_SPLINE = 3;

    //dane serializowane
    float data[];                   //curve samples
    float minx, maxx, miny, maxy;   //extents
	int samples;					//ilosc probek z krzywej
    String name;					//nazwa krzywej

    //dane tymczasowe - nieserializowane
    float m_afLastArg[];            // Tablica przechowujaca wartosci 
                                    //argumentow z ostatnich wywolan Eval

    Curve()
    {
        minx = miny = -1.0f;
        maxx = maxy = 1.0f;
        samples = 10;
        name = "curve";
        data = null;
        m_afLastArg = new float[1];
    }


    //samples curve and saves this data for future use
    void Build(int samples)
    {
        data = new float[samples];
		float step = (maxx - minx) / samples;
/*
        GameObject.Log("knots:");
        for (int j=0; j<n_knots; j++)
        {
            GameObject.Log("("+knots[j*2]+","+(knots[j*2]+1)+") ");
        }
        GameObject.LogLine("");
        */
        for (int i=0; i<samples; i++)
		{
			data[i] = Sample(i*step + minx);
		}
    }

    //evaluates curve at given point using precomputed values
    float Eval(float x)
    {
        if (x <= minx) return data[0];
        if (x >= maxx) return data[samples-1];

        float r = x - minx;
        int i = (int)((samples-1) * r / (maxx - minx));
        float f = r - (int)r;
        return data[i] * (1-f) + data[i+1] * f;
    }

	float GetMaxX() {return maxx;}
	float GetMinX() {return minx;}

    float Eval(float x, int id)
    {
		float result = Eval(x);

		if (m_afLastArg.length-1 < id)
        {
            m_afLastArg = new float[id+1];
        }
        
        m_afLastArg[id] = x;

        return result;
	}

    //--------------------------------------------------------
	//overridables
    float Sample(float x)
    {
		GameObject.CrashLog("Attempt to sample abstract curve\n");		
        return 0.0f;
    }

	void Read(DataInputStream s)
	{

	}

	void Write(DataOutputStream s)
	{

	}
    //---------------------------------------------------------
    
    byte GetType()
    {
        return CLASS_NONE;
    }
    
    public String toString()
    {
        return "ABSTRACT CURVE";
    }
}

/******************************************************************************
 * Implementuje krzywe typu spline
 */
class CurveSpline extends Curve
{
	float knots[];
	int n_knots;


    CurveSpline()
    {
        n_knots = 0;
        AddKnot(-1.0f,  0.0f);

        AddKnot(-0.5f,  0.5f);
        AddKnot(-.3f, .8f);
        AddKnot(.3f, -.8f);
        AddKnot( 0.5f, -0.5f);

        AddKnot( 1.0f,  0.0f);
    }

    void AddKnot(float x, float y)
    {
        //enlarge knots array
        if ( (knots == null) || ( (knots.length / 2) <= n_knots) )
        {
            float new_knots[] = new float[knots == null ? 8 : (knots.length + 8)];
            for (int i=0; i < n_knots*2; i++)
                new_knots[i] = knots[i];
            knots = new_knots;
        }

        //sort knots after x cordinate
        //look for 1st larger x than the new
        int i = 0;
        while ( ( i < n_knots) && (x > knots[2*i]) )
            i++;

        //GameObject.LogLine("1st larger: "+i);
        if (i < n_knots)
        {
            for (int j=n_knots; j > i; j--)
            {
                knots[ j*2 ] = knots[(j-1)*2];
                knots[ j*2+1 ] = knots[(j-1)*2+1];
            }
        }

        knots[i*2] = x;
        knots[i*2+1] = y;

        n_knots++;                        
    }

    void DeleteKnot(int k)
    {
        if ( (k >=0) && (k< n_knots) && (n_knots > 2) )
        {
            for (int i=k; i<n_knots-1; i++)
            {
                knots[i*2] = knots[(i+1)*2];
                knots[i*2+1] = knots[(i+1)*2+1];
            }
            n_knots--;

            float new_knots[] = new float[ n_knots*2 ];
            for (int j=0; j<n_knots; j++)
            {
                new_knots[j*2] = knots[j*2];
                new_knots[j*2+1] = knots[j*2+1];
            }
            knots = new_knots;
        }
    }

    byte GetType()
    {
        return CLASS_SPLINE;
    }

	float Sample(float x)
	{
		//look for index in knots table
		int i = 0;
        //GameObject.LogLine("computing sample:");
        /*for (int j=0; j<n_knots; j++)
        {
            GameObject.Log(""+j+":"+knots[j*2]+", "+knots[j*2+1]+"\n");
        }*/

        //first knot lying after argument x
		while ( (i < n_knots) && (knots[i*2] < x) )
			i++;
        //GameObject.LogLine("x:"+x+" knot index:"+i+" : "+knots[i*2]);

        //argument lying before first knot
        if (i == 0) return knots[0+1];
        //argument lying after last knot - return value of last knot
        if (i > n_knots-1) return knots[(n_knots-1)*2+1];

        //values of knots of interest, i.e. -2, -1, 0, 1
        float k0 = ((i-2) < 0) ? knots[0+1] : knots[(i-2)*2+1],
              k1 = knots[(i-1)*2+1],
              k2 = knots[i*2+1],
              k3 = ((i+1) > n_knots-1) ? knots[(n_knots-1)*2+1] : knots[(i+1)*2+1];

		//normalized x
		float xx = ( x - knots[(i-1)*2] ) / ( knots[i*2] - knots[(i-1)*2] );
        /*
        GameObject.LogLine("i:"+i+" x:"+x+" norm x:"+xx+" knots:"+
                           k0+ " " +
                           k1+ " " +
                           k2+ " " +
                           k3 );
        */

        float result = Tools.spline(xx, k0, k1, k2, k3);

        //GameObject.LogLine("\tresult:"+result);
		return result;

    }

    int GetClosestKnot(double x, double y)
    {
        int closest = -1;
        float dist = 1e10f;

        for (int i=0; i< n_knots; i++)
        {
            double d = Math.sqrt( Math.pow(x - knots[i*2], 2) + 
                                  Math.pow(y - knots[i*2+1], 2) );
            if (d < dist)
            {
                closest = i;
                dist = (float)d;
            }
        }
        
        return closest;
    }

    void SetKnotPos(int k, float x, float y)
    {
        if ((k >=0) && (k< n_knots))
        {
            knots[k*2] = x;
            knots[k*2+1] = y;
        }
    }
    
	void Read(DataInputStream s)
	{
		try
		{
			super.Read(s);
			n_knots = s.readUnsignedShort();

			if (n_knots < 2) 
			{
				GameObject.CrashLog("Error in spline curve: insufficient number of knots\n");
			}

			knots = new float[n_knots*2];

			//load 2D knots
			for (int i=0; i<n_knots*2; i++)
			{
				knots[i] = s.readFloat();
			}

		} catch (IOException e) 
		{
			GameObject.CrashLog("Error reading curve data: curve spline\n");
		}
	}

	void Write(DataOutputStream s)
	{
		try
		{
			super.Write(s);
			s.writeShort(n_knots);

			//write all knots coordinates
			for (int i=0; i<n_knots*2; i++)
			{
				s.writeFloat(knots[i]);
			}

		} catch (IOException e)
		{
			GameObject.CrashLog("Error writing curve data: curve spline\n");
		}
	}

    public String toString()
    {
        return "SPLINE";
    }    
}

/*
class CurveCarSettingsSpline extends CurveSpline
{
	void Read(DataInputStream s)
	{
		try
		{
			super.Read(s);
            
            s.readUTF();

		} catch (IOException e) 
		{
			GameObject.CrashLog("Error reading curve data: curve spline\n");
		}
    }
    
    void Write(DataOutputStream s)
	{
    
        //convert any CurveCarSettingsSpline objects into CurveSpline
        super.Write(s);

	}
}
*/