/** Trigger uruchamiajacy podane akcje (aCyclicAction) cyklicznie,
 * co czas pomiedzy fMinCycleTime i fMaxCycleTime. Dzialanie triggera
 * konczy sie gdy zostanie spelniony warunek komca (cEndCondition) lub
 * wykona sie nCyclicTimes cykli. Gdy nCyclicTimes==-1 to w nieskonczonosc
 * wolane sa cykle.
 * Gdy nCyclicTimes cykli jest wykonane to aEndActions wola sie po wykonaniu
 * jeszcze jednego cyklu, czyli z opoznieniem.
 */
public class TriggerCyclic extends GameObject implements iact_EnableDisable
{
    /** Znacznik, ze ten trigger aktualnie dziala */
    public boolean bEnabled = true;//#\TRIGGER
    
    /** Czas co ktory bedzie sprawdzany Trigger */
    public float fCheckInterval = 0.1f;//#\TRIGGER
    
    // akcje wykonywane cyklicznie
    public Action[] aCyclicActions = null;//#\CYCLIC
    
    // ile razy ma sie wykonac cykliczna akcja
    public int nCyclicTimes = -1;//#\CYCLIC

    // gdy warunek spelniony to konczy sie dzialanie
    public Condition cEndCondition = null;//#\TRIGGER

    // akcje wolane na koniec gdy warunek cEndCondition jest TRUE
    // lub gdy wykona sie nCyclicTimes cykli
    public Action[] aEndActions = null;//#\TRIGGER

    // minimalny czas cyklu
    public float fMinCycleTime = 0.0f;//#\CYCLIC
    
    // maksymalny czas cyklu
    public float fMaxCycleTime = 0.0f;//#\CYCLIC

    int nTimes = 0;
    
    public void OnCreate()
    {
		bSG = true;
		
        if (bEnabled)
            CallMethodIn("Enable", 0.0f);
    }
    
    public void OnDestroy()
    {
        if(cEndCondition != null)
        {
            cEndCondition.delete();
            cEndCondition = null;
        }
    }
    
    void Check()
    {
        if(!bEnabled)
        {
            DisableCallMethods();
            return;
        }
        
        // sprawdzamy warunek konca
        if(cEndCondition!=null && cEndCondition.Evaluate())
        {
            // wolamy akcje koncowe
            InvokeActionChain(aEndActions);
            
            // wylaczamy sie
            Disable();
        }
    }
    
    void Cycle()
    {
        // wylaczony trigger nie dziala
        if(!bEnabled)
        {
            DisableCallMethods();
            return;
        }
        
        // sprawdzamy warunek konca
        if((nTimes == 0) ||
           (cEndCondition!=null &&
            cEndCondition.PreEvaluate() &&
            cEndCondition.Evaluate()))
        {
            // wolamy akcje koncowe
            InvokeActionChain(aEndActions);
            
            // wylaczamy sie
            Disable();
            
            return;
        }
        
        // wlaczamy nastepny cykl
        if(nTimes != 0)
        {
            InvokeActionChain(aCyclicActions);
            if(nTimes>0) nTimes--;
        }
        
        CallNextCycle();
    }

    void CallNextCycle()
    {
        float fNextTime = fMinCycleTime + RandomFloat(fMaxCycleTime - fMinCycleTime);
        CallMethodIn("Cycle", fNextTime);
    }
    
    /** Metoda wolana po przyjsciu zdarzenia act_Enable */
    public void Enable()
    {
        bEnabled = true;
        
        nTimes = nCyclicTimes;
        
        // zerujemy warunek konca
        if(cEndCondition != null) cEndCondition.Clear();
        
        // wlaczamy odpalanie akcji
        CallMethodEvery("Check", fCheckInterval);
        CallNextCycle();
    }
    
    /** Metoda wolana po przyjsciu zdarzenia act_Disable */
    public void Disable()
    {
        bEnabled = false;
        DisableCallMethods();
    }
    
    /** Metoda sprawdza czy obiekt jest wlaczony czy wylaczony */
    public boolean IsEnabled()
    {
        return bEnabled;
    }
	
    //*********************************************************
	//					SaveGame
    //*********************************************************
	
    /** Metoda wczytujaca zapisane dane w Save gracza
     */	
    public void SGLoadChunk(FileChunk cParentFC)
    {
		FileChunk.Log("[TriggerCyclic.SGLoadChunk]\n");
		
		FileChunk cFC = GameObject.LoadChunk(cParentFC);
      
        if (cFC.GetID() == ESGChunks._CHK_TRIGGER_CYCLIC)
        {
			super.SGLoadChunk(cFC);
			
			bEnabled = cFC.LoadBool();
            
            cEndCondition = (Condition)cFC.LoadObject(cEndCondition,this);
            nTimes = cFC.LoadInt();
        }
		
		cFC.delete();
	}
	
    /** Metoda zapisujaca dane do Save gracza
     */		
    public void SGSaveChunk(FileChunk cFCParent)
    {
		FileChunk.Log("[TriggerCyclic.SGSaveChunk]\n");
		
        FileChunk cFC = NewChunk(ESGChunks._CHK_TRIGGER_CYCLIC, cFCParent);
		
		super.SGSaveChunk(cFC);

        cFC.SaveBool( bEnabled );
        
		cFC.SaveObject(cEndCondition);
        cFC.SaveInt( nTimes );
        
		cFC.delete();
	}	
	
	public void SGUpdate()
	{
		FileChunk.Log("[TriggerCyclic.SGUpdate]\n");
		super.SGUpdate();
		
        // wylaczony trigger nie dziala
        if(!bEnabled)
        {
            DisableCallMethods();
            return;
        }
	}
}
